"""DatasetMetadata <-> proto converters."""

from __future__ import annotations

from typing import TYPE_CHECKING

from radiens_core._converters._channels import channel_metadata_from_proto
from radiens_core._converters._enums import file_type_from_proto
from radiens_core.models.dataset import DatasetMetadata
from radiens_core.models.time_range import TimeRange

if TYPE_CHECKING:
    from radiens_core._generated import datasource_pb2


def dataset_metadata_from_proto(
    raw: datasource_pb2.DataSourceSetSaveReply,
) -> DatasetMetadata:
    """Build domain DatasetMetadata from proto DataSourceSetSaveReply."""
    status = raw.status
    sig_group = status.signalGroup
    pb_tr = status.tR

    channel_metadata = channel_metadata_from_proto(sig_group)

    # Build TimeRange from the proto TimeRange message
    sec_list = list(pb_tr.timeRangeSec)
    ts_list = list(pb_tr.timestamp)
    sec_start = sec_list[0] if len(sec_list) > 0 else 0.0
    sec_end = sec_list[1] if len(sec_list) > 1 else 0.0
    ts_start = ts_list[0] if len(ts_list) > 0 else 0
    ts_end = ts_list[1] if len(ts_list) > 1 else 0

    time_range = TimeRange(
        sec=(sec_start, sec_end),
        timestamp=(ts_start, ts_end),
        fs=pb_tr.fs,
        dur_sec=sec_end - sec_start,
        walltime=pb_tr.wallTimeStart,
        n=ts_end - ts_start,
    )

    return DatasetMetadata(
        id=raw.dsourceID,
        base_name=status.baseFileName,
        path=status.path,
        file_type=file_type_from_proto(status.fileType),
        time_range=time_range,
        channel_metadata=channel_metadata,
        parent_dsource_id="",
        probe_uid=sig_group.datasetUID,
    )



"""TimeRange <-> proto converters."""

from __future__ import annotations

from radiens_core._generated import common_pb2
from radiens_core.models.time_range import TimeRange


def time_range_from_proto(
    sec_start: float,
    sec_end: float,
    ts_start: int,
    ts_end: int,
    fs: float,
    dur_sec: float,
    walltime: str,
    n: int,
) -> TimeRange:
    """Build a domain TimeRange from proto fields."""
    return TimeRange(
        sec=(sec_start, sec_end),
        timestamp=(ts_start, ts_end),
        fs=fs,
        dur_sec=dur_sec,
        walltime=walltime,
        n=n,
    )


def time_range_to_proto(tr: TimeRange) -> common_pb2.TimeRange:
    """Convert domain TimeRange to proto TimeRange."""
    return common_pb2.TimeRange(
        timestamp=list(tr.timestamp),
        fs=tr.fs,
    )



"""VidereClient — offline file-based analysis client."""

from __future__ import annotations

from radiens_core._constants import ALLEGO_CORE_ADDR, DEFAULT_HUB_ID
from radiens_core._transport._auth import AuthInterceptor
from radiens_core._transport._channel_pool import ChannelPool
from radiens_core.models.dataset import DatasetMetadata
from radiens_core.models.enums import RadiensService


class VidereClient:
    """Client for offline file-based analysis.

    Usage::

        with VidereClient() as client:
            meta = client.link_data_file("/path/to/recording.xdat")
    """

    def __init__(self, address: str = ALLEGO_CORE_ADDR) -> None:
        self._address = address
        self._pool = ChannelPool(
            interceptors=[AuthInterceptor(server_address=address)],
        )
        self._datasets: dict[str, DatasetMetadata] = {}

    # --- Lifecycle ---

    def close(self) -> None:
        """Close all gRPC channels."""
        self._pool.close()

    def __enter__(self) -> VidereClient:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    # --- Internal helpers ---

    def _server_address(self, service: RadiensService) -> str:
        """Resolve server address for the given service."""
        return self._address

    def _resolve_dataset(self, ref: str | DatasetMetadata) -> DatasetMetadata:
        """Resolve a dataset reference to DatasetMetadata."""
        if isinstance(ref, DatasetMetadata):
            return ref
        ref_str = str(ref)
        if ref_str in self._datasets:
            return self._datasets[ref_str]
        raise ValueError(f"Unknown dataset: {ref_str}. Link it first with link_data_file().")

    # --- Data file operations ---

    def link_data_file(
        self,
        path: str,
        hub_name: str = DEFAULT_HUB_ID,
    ) -> DatasetMetadata:
        """Link a recording file and return its metadata.

        Parameters
        ----------
        path
            Path to the recording file (e.g., .xdat, .rhd).
        hub_name
            Hub name (default: "default").

        Returns
        -------
        DatasetMetadata
            Metadata for the linked dataset.
        """
        from radiens_core._services._radiens_core import set_datasource

        addr = self._server_address(RadiensService.CORE)
        meta = set_datasource(self._pool, addr, path)
        self._datasets[meta.id] = meta
        return meta



"""Signal selection and signal data array models."""

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, ConfigDict

from radiens_core.models.enums import KeyIndex


class SigSelect(BaseModel):
    """Signal selection specifying channels per signal type.

    Each field is a numpy array of native channel indices for that signal type.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    amp: npt.NDArray[np.int64]
    gpio_ain: npt.NDArray[np.int64]
    gpio_din: npt.NDArray[np.int64]
    gpio_dout: npt.NDArray[np.int64]
    key_idx: KeyIndex


class SignalArrays(BaseModel):
    """Container for signal data arrays, one per signal type.

    Each field is either a 2D numpy array ``(channels, samples)``
    or ``None`` if that signal type was not requested.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    amp: npt.NDArray[np.float64] | None = None
    gpio_ain: npt.NDArray[np.float64] | None = None
    gpio_din: npt.NDArray[np.float64] | None = None
    gpio_dout: npt.NDArray[np.float64] | None = None



"""CurateClient — data curation client (skeleton)."""

from __future__ import annotations

from typing import TYPE_CHECKING

from radiens_core._constants import ALLEGO_CORE_ADDR
from radiens_core._transport._auth import AuthInterceptor
from radiens_core._transport._channel_pool import ChannelPool

if TYPE_CHECKING:
    from radiens_core.models.enums import RadiensService


class CurateClient:
    """Client for data curation operations (skeleton)."""

    def __init__(self, address: str = ALLEGO_CORE_ADDR) -> None:
        self._address = address
        self._pool = ChannelPool(
            interceptors=[AuthInterceptor(server_address=address)],
        )

    # --- Lifecycle ---

    def close(self) -> None:
        """Close all gRPC channels."""
        self._pool.close()

    def __enter__(self) -> CurateClient:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    # --- Internal helpers ---

    def _server_address(self, service: RadiensService) -> str:
        """Resolve server address for the given service."""
        return self._address



"""Request model pattern — used by public client methods.

Every non-trivial RPC provides:

1. A frozen Pydantic model with :data:`FlexTimeRange` / :data:`FlexSignalSpec`
   fields so raw Python types are coerced automatically.
2. An ``@overload``-ed client method that accepts either the model
   *or* individual kwargs, giving callers a choice between structured
   and ad-hoc invocations.

Example (for ``get_signals``, to be wired when that RPC is restored)::

    class GetSignalsRequest(BaseModel):
        model_config = ConfigDict(frozen=True)

        time_range: FlexTimeRange | None = None
        signal_spec: FlexSignalSpec | None = None
        dataset: str | DatasetMetadata | None = None
        hub_name: str = DEFAULT_HUB_ID

    # On AllegoClient:
    @overload
    def get_signals(self, request: GetSignalsRequest) -> SignalArrays: ...

    @overload
    def get_signals(
        self,
        *,
        time_range: TimeRangeSpec | list[float] | None = ...,
        signal_spec: SignalSpec | list[int] | dict[str, Any] | None = ...,
        dataset: str | DatasetMetadata | None = ...,
        hub_name: str = ...,
    ) -> SignalArrays: ...

    def get_signals(
        self,
        request: GetSignalsRequest | None = None,
        **kwargs: Any,
    ) -> SignalArrays:
        if request is None:
            request = GetSignalsRequest(**kwargs)
        # Pydantic has already normalized all inputs via BeforeValidator
        ...

This module is a *template reference* — it doesn't export anything used
at runtime. Concrete request models are defined next to the client method
that uses them (e.g., in ``allego_client.py``).
"""



"""Port domain models."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict

from radiens_core.models.enums import Port


class PortChannelCount(BaseModel):
    """Port channel count information."""

    model_config = ConfigDict(frozen=True)

    port: Port
    channel_count: int



"""Trigger control converters — domain ↔ proto."""

from __future__ import annotations

from typing import TYPE_CHECKING

from radiens_core._converters._enums import (
    port_from_proto,
    port_to_proto,
    trigger_channel_from_proto,
    trigger_channel_to_proto,
)
from radiens_core.models.status import (
    ManualStimTriggerRequest,
    ManualStimTriggerToggleRequest,
    SetTriggerStateRequest,
    TriggerState,
)

if TYPE_CHECKING:
    from radiens_core._generated import allegoserver_pb2


def set_trigger_state_request_to_proto(
    request: SetTriggerStateRequest,
) -> allegoserver_pb2.SetTriggerRequest:
    """Convert ``SetTriggerStateRequest`` domain model → proto message."""
    from radiens_core._generated import allegoserver_pb2

    return allegoserver_pb2.SetTriggerRequest(
        channel=trigger_channel_to_proto(request.channel),
        enabled=request.enabled,
        port=port_to_proto(request.port),
    )


def trigger_state_from_proto(
    response: allegoserver_pb2.TriggerState,
) -> TriggerState:
    """Convert ``TriggerState`` proto response → domain model."""
    return TriggerState(
        enabled_channels=[
            trigger_channel_from_proto(channel) for channel in response.enabledChannels
        ],
        ports=[port_from_proto(port) for port in response.ports],
    )


def manual_stim_trigger_request_to_proto(
    request: ManualStimTriggerRequest,
) -> allegoserver_pb2.ManualStimTriggerRequest:
    """Convert ``ManualStimTriggerRequest`` domain model → proto message."""
    from radiens_core._generated import allegoserver_pb2

    # StimKeypressIndex is 1-indexed (1-8); proto expects 0-indexed (0-7)
    return allegoserver_pb2.ManualStimTriggerRequest(
        trigger=request.trigger - 1,
        triggerOn=request.trigger_on,
    )


def manual_stim_trigger_toggle_request_to_proto(
    request: ManualStimTriggerToggleRequest,
) -> allegoserver_pb2.ManualStimTriggerToggleRequest:
    """Convert ``ManualStimTriggerToggleRequest`` domain model → proto message."""
    from radiens_core._generated import allegoserver_pb2

    # StimKeypressIndex is 1-indexed (1-8); proto expects 0-indexed (0-7)
    return allegoserver_pb2.ManualStimTriggerToggleRequest(
        trigger=request.trigger - 1,
    )



"""RadiensCore RPC wrappers — thin service layer (proto only)."""

from __future__ import annotations

from typing import TYPE_CHECKING

import grpc

from radiens_core._converters._enums import file_type_to_proto
from radiens_core._generated import datasource_pb2, radiensserver_pb2_grpc
from radiens_core._transport._errors import handle_grpc_error
from radiens_core.models.enums import ClientType, RadiensFileType

if TYPE_CHECKING:
    from radiens_core._transport._channel_pool import ChannelPool
    from radiens_core.models.dataset import DatasetMetadata


def _guess_file_type(path: str) -> RadiensFileType:
    """Guess file type from file extension."""
    ext = path.rsplit(".", maxsplit=1)[-1].lower() if "." in path else ""
    ext_map: dict[str, RadiensFileType] = {
        "xdat": RadiensFileType.XDAT,
        "rhd": RadiensFileType.RHD,
        "csv": RadiensFileType.CSV,
        "hdf5": RadiensFileType.HDF5,
        "h5": RadiensFileType.HDF5,
        "nex5": RadiensFileType.NEX5,
        "nwb": RadiensFileType.NWB,
        "bin": RadiensFileType.KILOSORT2,
        "nsx": RadiensFileType.NSX,
        "tdt": RadiensFileType.TDT,
        "oebin": RadiensFileType.OE_DAT,
    }
    return ext_map.get(ext, RadiensFileType.XDAT)


def set_datasource(
    pool: ChannelPool,
    addr: str,
    path: str,
) -> DatasetMetadata:
    """Link a data file via SetDataSourceFromFile RPC.

    Returns domain DatasetMetadata.
    """
    import os

    from radiens_core._converters._dataset import dataset_metadata_from_proto

    base_name = os.path.splitext(os.path.basename(path))[0]
    dir_path = os.path.dirname(path) or "."

    req = datasource_pb2.DataSourceSetSaveRequest(
        path=dir_path,
        baseName=base_name,
        fileType=file_type_to_proto(_guess_file_type(path)),
    )

    channel = pool.get_channel(addr)
    stub = radiensserver_pb2_grpc.RadiensCoreStub(channel)
    try:
        res = stub.SetDataSourceFromFile(req)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.VIDERE)
        raise  # unreachable, but satisfies mypy

    return dataset_metadata_from_proto(res)



"""Domain enum <-> proto enum mappings.

This is the ONLY bridge between domain enums and proto enums.
Adding a proto enum value without updating these maps -> mypy error via exhaustive match.
"""

from __future__ import annotations

from radiens_core._generated import allegoserver_pb2, common_pb2
from radiens_core.models.enums import (
    BackboneMode,
    DataRecordingType,
    Port,
    RadiensFileType,
    RecordMode,
    SignalType,
    StimPolarity,
    StimPulseMode,
    StimShape,
    StimStep,
    StimTriggerEdgeOrLevel,
    StimTriggerHighOrLow,
    StreamMode,
    TriggerChannel,
    TrsMode,
)

# --- SignalType ---

_SIGNAL_TYPE_TO_PROTO: dict[SignalType, common_pb2.SignalType.ValueType] = {
    SignalType.AMP: common_pb2.PRI,
    SignalType.AIN: common_pb2.AUX,
    SignalType.DIN: common_pb2.DIN,
    SignalType.DOUT: common_pb2.DOUT,
}

_SIGNAL_TYPE_FROM_PROTO: dict[int, SignalType] = {
    int(common_pb2.PRI): SignalType.AMP,
    int(common_pb2.AUX): SignalType.AIN,
    int(common_pb2.DIN): SignalType.DIN,
    int(common_pb2.DOUT): SignalType.DOUT,
}


def signal_type_to_proto(st: SignalType) -> common_pb2.SignalType.ValueType:
    """Convert domain SignalType to proto SignalType."""
    return _SIGNAL_TYPE_TO_PROTO[st]


def signal_type_from_proto(pb: int) -> SignalType:
    """Convert proto SignalType int to domain SignalType."""
    return _SIGNAL_TYPE_FROM_PROTO[pb]


# --- TrsMode ---

_TRS_MODE_TO_PROTO: dict[TrsMode, common_pb2.TimeRangeSelMode.ValueType] = {
    TrsMode.SUBSET: common_pb2.TRS_SUBSET,
    TrsMode.TO_HEAD: common_pb2.TRS_TO_HEAD,
    TrsMode.FROM_HEAD: common_pb2.TRS_FROM_HEAD,
}

_TRS_MODE_FROM_PROTO: dict[int, TrsMode] = {
    int(common_pb2.TRS_SUBSET): TrsMode.SUBSET,
    int(common_pb2.TRS_TO_HEAD): TrsMode.TO_HEAD,
    int(common_pb2.TRS_FROM_HEAD): TrsMode.FROM_HEAD,
}


def trs_mode_to_proto(mode: TrsMode) -> common_pb2.TimeRangeSelMode.ValueType:
    """Convert domain TrsMode to proto TimeRangeSelMode."""
    return _TRS_MODE_TO_PROTO[mode]


def trs_mode_from_proto(pb: int) -> TrsMode:
    """Convert proto TimeRangeSelMode int to domain TrsMode."""
    return _TRS_MODE_FROM_PROTO[pb]


# --- RadiensFileType ---

_FILE_TYPE_FROM_PROTO: dict[int, RadiensFileType] = {
    int(common_pb2.RHD): RadiensFileType.RHD,
    int(common_pb2.XDAT): RadiensFileType.XDAT,
    int(common_pb2.CSV): RadiensFileType.CSV,
    int(common_pb2.HDF5): RadiensFileType.HDF5,
    int(common_pb2.NEX5): RadiensFileType.NEX5,
    int(common_pb2.NWB): RadiensFileType.NWB,
    int(common_pb2.KILOSORT2): RadiensFileType.KILOSORT2,
    int(common_pb2.NSX): RadiensFileType.NSX,
    int(common_pb2.TDT): RadiensFileType.TDT,
    int(common_pb2.SPIKES): RadiensFileType.SPIKES,
    int(common_pb2.OE_DAT): RadiensFileType.OE_DAT,
}


def file_type_from_proto(pb: int) -> RadiensFileType:
    """Convert proto RadixFileTypes int to domain RadiensFileType."""
    return _FILE_TYPE_FROM_PROTO[pb]


_FILE_TYPE_TO_PROTO: dict[RadiensFileType, common_pb2.RadixFileTypes.ValueType] = {
    RadiensFileType.RHD: common_pb2.RHD,
    RadiensFileType.XDAT: common_pb2.XDAT,
    RadiensFileType.CSV: common_pb2.CSV,
    RadiensFileType.HDF5: common_pb2.HDF5,
    RadiensFileType.NEX5: common_pb2.NEX5,
    RadiensFileType.NWB: common_pb2.NWB,
    RadiensFileType.KILOSORT2: common_pb2.KILOSORT2,
    RadiensFileType.NSX: common_pb2.NSX,
    RadiensFileType.TDT: common_pb2.TDT,
    RadiensFileType.SPIKES: common_pb2.SPIKES,
    RadiensFileType.OE_DAT: common_pb2.OE_DAT,
}


def file_type_to_proto(ft: RadiensFileType) -> common_pb2.RadixFileTypes.ValueType:
    """Convert domain RadiensFileType to proto RadixFileTypes."""
    return _FILE_TYPE_TO_PROTO[ft]


# --- StimShape ---

_STIM_SHAPE_TO_PROTO: dict[StimShape, allegoserver_pb2.StimShape.ValueType] = {
    StimShape.BIPHASIC: allegoserver_pb2.BIPHASIC,
    StimShape.BIPHASIC_INTERPHASE_DELAY: allegoserver_pb2.BIPHASIC_INTERPHASE_DELAY,
    StimShape.TRIPHASIC: allegoserver_pb2.TRIPHASIC,
    StimShape.MONOPHASIC: allegoserver_pb2.MONOPHASIC,
}

_STIM_SHAPE_FROM_PROTO: dict[int, StimShape] = {
    int(allegoserver_pb2.BIPHASIC): StimShape.BIPHASIC,
    int(allegoserver_pb2.BIPHASIC_INTERPHASE_DELAY): StimShape.BIPHASIC_INTERPHASE_DELAY,
    int(allegoserver_pb2.TRIPHASIC): StimShape.TRIPHASIC,
    int(allegoserver_pb2.MONOPHASIC): StimShape.MONOPHASIC,
}


def stim_shape_to_proto(ss: StimShape) -> allegoserver_pb2.StimShape.ValueType:
    """Convert domain StimShape to proto StimShape."""
    return _STIM_SHAPE_TO_PROTO[ss]


def stim_shape_from_proto(pb: int) -> StimShape:
    """Convert proto StimShape int to domain StimShape."""
    return _STIM_SHAPE_FROM_PROTO[pb]


# --- StimPolarity ---

_STIM_POLARITY_TO_PROTO: dict[StimPolarity, allegoserver_pb2.StimPolarity.ValueType] = {
    StimPolarity.CATHODIC_FIRST: allegoserver_pb2.CATHODIC_FIRST,
    StimPolarity.ANODIC_FIRST: allegoserver_pb2.ANODIC_FIRST,
}

_STIM_POLARITY_FROM_PROTO: dict[int, StimPolarity] = {
    int(allegoserver_pb2.CATHODIC_FIRST): StimPolarity.CATHODIC_FIRST,
    int(allegoserver_pb2.ANODIC_FIRST): StimPolarity.ANODIC_FIRST,
}


def stim_polarity_to_proto(sp: StimPolarity) -> allegoserver_pb2.StimPolarity.ValueType:
    """Convert domain StimPolarity to proto StimPolarity."""
    return _STIM_POLARITY_TO_PROTO[sp]


def stim_polarity_from_proto(pb: int) -> StimPolarity:
    """Convert proto StimPolarity int to domain StimPolarity."""
    return _STIM_POLARITY_FROM_PROTO[pb]


# --- StimTriggerEdgeOrLevel ---

_STIM_TRIGGER_EDGE_OR_LEVEL_TO_PROTO: dict[
    StimTriggerEdgeOrLevel, allegoserver_pb2.StimTriggerEdgeOrLevel.ValueType
] = {
    StimTriggerEdgeOrLevel.EDGE: allegoserver_pb2.ST_EDGE,
    StimTriggerEdgeOrLevel.LEVEL: allegoserver_pb2.ST_LEVEL,
}

_STIM_TRIGGER_EDGE_OR_LEVEL_FROM_PROTO: dict[int, StimTriggerEdgeOrLevel] = {
    int(allegoserver_pb2.ST_EDGE): StimTriggerEdgeOrLevel.EDGE,
    int(allegoserver_pb2.ST_LEVEL): StimTriggerEdgeOrLevel.LEVEL,
}


def stim_trigger_edge_or_level_to_proto(
    ste: StimTriggerEdgeOrLevel,
) -> allegoserver_pb2.StimTriggerEdgeOrLevel.ValueType:
    """Convert domain StimTriggerEdgeOrLevel to proto StimTriggerEdgeOrLevel."""
    return _STIM_TRIGGER_EDGE_OR_LEVEL_TO_PROTO[ste]


def stim_trigger_edge_or_level_from_proto(pb: int) -> StimTriggerEdgeOrLevel:
    """Convert proto StimTriggerEdgeOrLevel int to domain StimTriggerEdgeOrLevel."""
    return _STIM_TRIGGER_EDGE_OR_LEVEL_FROM_PROTO[pb]


# --- StimTriggerHighOrLow ---

_STIM_TRIGGER_HIGH_OR_LOW_TO_PROTO: dict[StimTriggerHighOrLow, allegoserver_pb2.StimTriggerHighOrLow.ValueType] = {
    StimTriggerHighOrLow.HIGH: allegoserver_pb2.ST_HIGH,
    StimTriggerHighOrLow.LOW: allegoserver_pb2.ST_LOW,
}

_STIM_TRIGGER_HIGH_OR_LOW_FROM_PROTO: dict[int, StimTriggerHighOrLow] = {
    int(allegoserver_pb2.ST_HIGH): StimTriggerHighOrLow.HIGH,
    int(allegoserver_pb2.ST_LOW): StimTriggerHighOrLow.LOW,
}


def stim_trigger_level_to_proto(
    stl: StimTriggerHighOrLow,
) -> allegoserver_pb2.StimTriggerHighOrLow.ValueType:
    """Convert domain StimTriggerLevel to proto StimTriggerHighOrLow."""
    return _STIM_TRIGGER_HIGH_OR_LOW_TO_PROTO[stl]


def stim_trigger_level_from_proto(pb: int) -> StimTriggerHighOrLow:
    """Convert proto StimTriggerHighOrLow int to domain StimTriggerLevel."""
    return _STIM_TRIGGER_HIGH_OR_LOW_FROM_PROTO[pb]


# --- StimPulseMode ---

_STIM_PULSE_MODE_TO_PROTO: dict[StimPulseMode, allegoserver_pb2.StimPulseOrTrain.ValueType] = {
    StimPulseMode.SINGLE_PULSE: allegoserver_pb2.SINGLE_PULSE,
    StimPulseMode.PULSE_TRAIN: allegoserver_pb2.PULSE_TRAIN,
}

_STIM_PULSE_MODE_FROM_PROTO: dict[int, StimPulseMode] = {
    int(allegoserver_pb2.SINGLE_PULSE): StimPulseMode.SINGLE_PULSE,
    int(allegoserver_pb2.PULSE_TRAIN): StimPulseMode.PULSE_TRAIN,
}


def stim_pulse_mode_to_proto(
    spm: StimPulseMode,
) -> allegoserver_pb2.StimPulseOrTrain.ValueType:
    """Convert domain StimPulseMode to proto StimPulseOrTrain."""
    return _STIM_PULSE_MODE_TO_PROTO[spm]


def stim_pulse_mode_from_proto(pb: int) -> StimPulseMode:
    """Convert proto StimPulseOrTrain int to domain StimPulseMode."""
    return _STIM_PULSE_MODE_FROM_PROTO[pb]


# --- StreamMode ---

_STREAM_MODE_TO_PROTO: dict[StreamMode, allegoserver_pb2.StreamMode.ValueType] = {
    StreamMode.OFF: allegoserver_pb2.S_OFF,
    StreamMode.ON: allegoserver_pb2.S_ON,
}

_STREAM_MODE_FROM_PROTO: dict[int, StreamMode] = {
    int(allegoserver_pb2.S_OFF): StreamMode.OFF,
    int(allegoserver_pb2.S_ON): StreamMode.ON,
}


def stream_mode_to_proto(sm: StreamMode) -> allegoserver_pb2.StreamMode.ValueType:
    """Convert domain StreamMode to proto StreamMode."""
    return _STREAM_MODE_TO_PROTO[sm]


def stream_mode_from_proto(pb: int) -> StreamMode:
    """Convert proto StreamMode int to domain StreamMode."""
    return _STREAM_MODE_FROM_PROTO[pb]


# --- RecordMode ---

_RECORD_MODE_TO_PROTO: dict[RecordMode, allegoserver_pb2.RecordMode.ValueType] = {
    RecordMode.OFF: allegoserver_pb2.R_OFF,
    RecordMode.ON: allegoserver_pb2.R_ON,
}

_RECORD_MODE_FROM_PROTO: dict[int, RecordMode] = {
    int(allegoserver_pb2.R_OFF): RecordMode.OFF,
    int(allegoserver_pb2.R_ON): RecordMode.ON,
}


def record_mode_to_proto(rm: RecordMode) -> allegoserver_pb2.RecordMode.ValueType:
    """Convert domain RecordMode to proto RecordMode."""
    return _RECORD_MODE_TO_PROTO[rm]


def record_mode_from_proto(pb: int) -> RecordMode:
    """Convert proto RecordMode int to domain RecordMode."""
    return _RECORD_MODE_FROM_PROTO[pb]


# --- Port ---

_PORT_TO_PROTO: dict[Port, common_pb2.Port.ValueType] = {
    Port.A: common_pb2.A,
    Port.B: common_pb2.B,
    Port.C: common_pb2.C,
    Port.D: common_pb2.D,
    Port.E: common_pb2.E,
    Port.F: common_pb2.F,
    Port.G: common_pb2.G,
    Port.H: common_pb2.H,
}

_PORT_FROM_PROTO: dict[int, Port] = {
    int(common_pb2.A): Port.A,
    int(common_pb2.B): Port.B,
    int(common_pb2.C): Port.C,
    int(common_pb2.D): Port.D,
    int(common_pb2.E): Port.E,
    int(common_pb2.F): Port.F,
    int(common_pb2.G): Port.G,
    int(common_pb2.H): Port.H,
}


def port_to_proto(port: Port) -> common_pb2.Port.ValueType:
    """Convert domain Port to proto Port."""
    return _PORT_TO_PROTO[port]


def port_from_proto(pb: int) -> Port:
    """Convert proto Port int to domain Port."""
    return _PORT_FROM_PROTO[pb]


# --- BackboneMode ---

_BACKBONE_MODE_TO_PROTO: dict[BackboneMode, allegoserver_pb2.BackboneMode.ValueType] = {
    BackboneMode.SMARTBOX_PRO: allegoserver_pb2.SMARTBOX_PRO,
    BackboneMode.SMARTBOX_SIM_GEN_SINE: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE,
    BackboneMode.SMARTBOX_SIM_GEN_SPIKES: allegoserver_pb2.SMARTBOX_SIM_GEN_SPIKES,
    BackboneMode.SMARTBOX_SIM_DATASOURCE: allegoserver_pb2.SMARTBOX_SIM_DATASOURCE,
    BackboneMode.OPEN_EPHYS_USB3: allegoserver_pb2.OPEN_EPHYS_USB3,
    BackboneMode.INTAN_RECORDING_CONTROLLER_1024: allegoserver_pb2.INTAN_RECORDING_CONTROLLER_1024,
    BackboneMode.SMARTBOX_CLASSIC: allegoserver_pb2.SMARTBOX_CLASSIC,
    BackboneMode.INTAN_RECORDING_CONTROLLER_512: allegoserver_pb2.INTAN_RECORDING_CONTROLLER_512,
    BackboneMode.OPEN_EPHYS_USB2: allegoserver_pb2.OPEN_EPHYS_USB2,
    BackboneMode.INTAN_USB2: allegoserver_pb2.INTAN_USB2,
    BackboneMode.SMARTBOX_SIM_GEN_SINE_MAPPED: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MAPPED,
    BackboneMode.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ,
    BackboneMode.SMARTBOX_SIM_GEN_SINE_MULTI_BAND: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MULTI_BAND,
    BackboneMode.SMARTBOX_PRO_SINAPS_256: allegoserver_pb2.SMARTBOX_PRO_SINAPS_256,
    BackboneMode.SMARTBOX_PRO_SINAPS_1024: allegoserver_pb2.SMARTBOX_PRO_SINAPS_1024,
    BackboneMode.XDAQ_ONE_REC: allegoserver_pb2.XDAQ_ONE_REC,
    BackboneMode.XDAQ_ONE_STIM: allegoserver_pb2.XDAQ_ONE_STIM,
    BackboneMode.XDAQ_CORE_REC: allegoserver_pb2.XDAQ_CORE_REC,
    BackboneMode.XDAQ_CORE_STIM: allegoserver_pb2.XDAQ_CORE_STIM,
    BackboneMode.SMARTBOX_PRO_SINAPS_512: allegoserver_pb2.SMARTBOX_PRO_SINAPS_512,
    BackboneMode.XDAQ_GEN2_ONE_STIM: allegoserver_pb2.XDAQ_GEN2_ONE_STIM,
    BackboneMode.XDAQ_GEN2_CORE_STIM: allegoserver_pb2.XDAQ_GEN2_CORE_STIM,
    BackboneMode.XDAQ_GEN2_ONE_REC: allegoserver_pb2.XDAQ_GEN2_ONE_REC,
    BackboneMode.XDAQ_GEN2_CORE_REC: allegoserver_pb2.XDAQ_GEN2_CORE_REC,
    BackboneMode.SMARTBOX_SIM_GEN_LFP: allegoserver_pb2.SMARTBOX_SIM_GEN_LFP,
}

_BACKBONE_MODE_FROM_PROTO: dict[int, BackboneMode] = {
    int(allegoserver_pb2.SMARTBOX_PRO): BackboneMode.SMARTBOX_PRO,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE): BackboneMode.SMARTBOX_SIM_GEN_SINE,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SPIKES): BackboneMode.SMARTBOX_SIM_GEN_SPIKES,
    int(allegoserver_pb2.SMARTBOX_SIM_DATASOURCE): BackboneMode.SMARTBOX_SIM_DATASOURCE,
    int(allegoserver_pb2.OPEN_EPHYS_USB3): BackboneMode.OPEN_EPHYS_USB3,
    int(allegoserver_pb2.INTAN_RECORDING_CONTROLLER_1024): BackboneMode.INTAN_RECORDING_CONTROLLER_1024,
    int(allegoserver_pb2.SMARTBOX_CLASSIC): BackboneMode.SMARTBOX_CLASSIC,
    int(allegoserver_pb2.INTAN_RECORDING_CONTROLLER_512): BackboneMode.INTAN_RECORDING_CONTROLLER_512,
    int(allegoserver_pb2.OPEN_EPHYS_USB2): BackboneMode.OPEN_EPHYS_USB2,
    int(allegoserver_pb2.INTAN_USB2): BackboneMode.INTAN_USB2,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MAPPED): BackboneMode.SMARTBOX_SIM_GEN_SINE_MAPPED,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ): BackboneMode.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MULTI_BAND): BackboneMode.SMARTBOX_SIM_GEN_SINE_MULTI_BAND,
    int(allegoserver_pb2.SMARTBOX_PRO_SINAPS_256): BackboneMode.SMARTBOX_PRO_SINAPS_256,
    int(allegoserver_pb2.SMARTBOX_PRO_SINAPS_1024): BackboneMode.SMARTBOX_PRO_SINAPS_1024,
    int(allegoserver_pb2.XDAQ_ONE_REC): BackboneMode.XDAQ_ONE_REC,
    int(allegoserver_pb2.XDAQ_ONE_STIM): BackboneMode.XDAQ_ONE_STIM,
    int(allegoserver_pb2.XDAQ_CORE_REC): BackboneMode.XDAQ_CORE_REC,
    int(allegoserver_pb2.XDAQ_CORE_STIM): BackboneMode.XDAQ_CORE_STIM,
    int(allegoserver_pb2.SMARTBOX_PRO_SINAPS_512): BackboneMode.SMARTBOX_PRO_SINAPS_512,
    int(allegoserver_pb2.XDAQ_GEN2_ONE_STIM): BackboneMode.XDAQ_GEN2_ONE_STIM,
    int(allegoserver_pb2.XDAQ_GEN2_CORE_STIM): BackboneMode.XDAQ_GEN2_CORE_STIM,
    int(allegoserver_pb2.XDAQ_GEN2_ONE_REC): BackboneMode.XDAQ_GEN2_ONE_REC,
    int(allegoserver_pb2.XDAQ_GEN2_CORE_REC): BackboneMode.XDAQ_GEN2_CORE_REC,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_LFP): BackboneMode.SMARTBOX_SIM_GEN_LFP,
}


def backbone_mode_to_proto(bm: BackboneMode) -> allegoserver_pb2.BackboneMode.ValueType:
    """Convert domain BackboneMode to proto BackboneMode."""
    return _BACKBONE_MODE_TO_PROTO[bm]


def backbone_mode_from_proto(pb: int) -> BackboneMode:
    """Convert proto BackboneMode int to domain BackboneMode."""
    return _BACKBONE_MODE_FROM_PROTO[pb]


# --- TriggerChannel ---

_TRIGGER_CHANNEL_TO_PROTO: dict[TriggerChannel, allegoserver_pb2.TriggerChannel.ValueType] = {
    TriggerChannel.T_DIN_0: allegoserver_pb2.T_DIN_0,
    TriggerChannel.T_DIN_1: allegoserver_pb2.T_DIN_1,
    TriggerChannel.T_DIN_2: allegoserver_pb2.T_DIN_2,
    TriggerChannel.T_DIN_3: allegoserver_pb2.T_DIN_3,
    TriggerChannel.T_DIN_4: allegoserver_pb2.T_DIN_4,
    TriggerChannel.T_DIN_5: allegoserver_pb2.T_DIN_5,
    TriggerChannel.T_DIN_6: allegoserver_pb2.T_DIN_6,
    TriggerChannel.T_DIN_7: allegoserver_pb2.T_DIN_7,
    TriggerChannel.T_DIN_8: allegoserver_pb2.T_DIN_8,
    TriggerChannel.T_DIN_9: allegoserver_pb2.T_DIN_9,
    TriggerChannel.T_DIN_10: allegoserver_pb2.T_DIN_10,
    TriggerChannel.T_DIN_11: allegoserver_pb2.T_DIN_11,
    TriggerChannel.T_DIN_12: allegoserver_pb2.T_DIN_12,
    TriggerChannel.T_DIN_13: allegoserver_pb2.T_DIN_13,
    TriggerChannel.T_DIN_14: allegoserver_pb2.T_DIN_14,
    TriggerChannel.T_DIN_15: allegoserver_pb2.T_DIN_15,
    TriggerChannel.T_DOUT_0: allegoserver_pb2.T_DOUT_0,
    TriggerChannel.T_DOUT_1: allegoserver_pb2.T_DOUT_1,
    TriggerChannel.T_DOUT_2: allegoserver_pb2.T_DOUT_2,
    TriggerChannel.T_DOUT_3: allegoserver_pb2.T_DOUT_3,
    TriggerChannel.T_DOUT_4: allegoserver_pb2.T_DOUT_4,
    TriggerChannel.T_DOUT_5: allegoserver_pb2.T_DOUT_5,
    TriggerChannel.T_DOUT_6: allegoserver_pb2.T_DOUT_6,
    TriggerChannel.T_DOUT_7: allegoserver_pb2.T_DOUT_7,
    TriggerChannel.T_DOUT_8: allegoserver_pb2.T_DOUT_8,
    TriggerChannel.T_DOUT_9: allegoserver_pb2.T_DOUT_9,
    TriggerChannel.T_DOUT_10: allegoserver_pb2.T_DOUT_10,
    TriggerChannel.T_DOUT_11: allegoserver_pb2.T_DOUT_11,
    TriggerChannel.T_DOUT_12: allegoserver_pb2.T_DOUT_12,
    TriggerChannel.T_DOUT_13: allegoserver_pb2.T_DOUT_13,
    TriggerChannel.T_DOUT_14: allegoserver_pb2.T_DOUT_14,
    TriggerChannel.T_DOUT_15: allegoserver_pb2.T_DOUT_15,
    TriggerChannel.T_NONE: allegoserver_pb2.T_NONE,
}

_TRIGGER_CHANNEL_FROM_PROTO: dict[int, TriggerChannel] = {
    int(allegoserver_pb2.T_DIN_0): TriggerChannel.T_DIN_0,
    int(allegoserver_pb2.T_DIN_1): TriggerChannel.T_DIN_1,
    int(allegoserver_pb2.T_DIN_2): TriggerChannel.T_DIN_2,
    int(allegoserver_pb2.T_DIN_3): TriggerChannel.T_DIN_3,
    int(allegoserver_pb2.T_DIN_4): TriggerChannel.T_DIN_4,
    int(allegoserver_pb2.T_DIN_5): TriggerChannel.T_DIN_5,
    int(allegoserver_pb2.T_DIN_6): TriggerChannel.T_DIN_6,
    int(allegoserver_pb2.T_DIN_7): TriggerChannel.T_DIN_7,
    int(allegoserver_pb2.T_DIN_8): TriggerChannel.T_DIN_8,
    int(allegoserver_pb2.T_DIN_9): TriggerChannel.T_DIN_9,
    int(allegoserver_pb2.T_DIN_10): TriggerChannel.T_DIN_10,
    int(allegoserver_pb2.T_DIN_11): TriggerChannel.T_DIN_11,
    int(allegoserver_pb2.T_DIN_12): TriggerChannel.T_DIN_12,
    int(allegoserver_pb2.T_DIN_13): TriggerChannel.T_DIN_13,
    int(allegoserver_pb2.T_DIN_14): TriggerChannel.T_DIN_14,
    int(allegoserver_pb2.T_DIN_15): TriggerChannel.T_DIN_15,
    int(allegoserver_pb2.T_DOUT_0): TriggerChannel.T_DOUT_0,
    int(allegoserver_pb2.T_DOUT_1): TriggerChannel.T_DOUT_1,
    int(allegoserver_pb2.T_DOUT_2): TriggerChannel.T_DOUT_2,
    int(allegoserver_pb2.T_DOUT_3): TriggerChannel.T_DOUT_3,
    int(allegoserver_pb2.T_DOUT_4): TriggerChannel.T_DOUT_4,
    int(allegoserver_pb2.T_DOUT_5): TriggerChannel.T_DOUT_5,
    int(allegoserver_pb2.T_DOUT_6): TriggerChannel.T_DOUT_6,
    int(allegoserver_pb2.T_DOUT_7): TriggerChannel.T_DOUT_7,
    int(allegoserver_pb2.T_DOUT_8): TriggerChannel.T_DOUT_8,
    int(allegoserver_pb2.T_DOUT_9): TriggerChannel.T_DOUT_9,
    int(allegoserver_pb2.T_DOUT_10): TriggerChannel.T_DOUT_10,
    int(allegoserver_pb2.T_DOUT_11): TriggerChannel.T_DOUT_11,
    int(allegoserver_pb2.T_DOUT_12): TriggerChannel.T_DOUT_12,
    int(allegoserver_pb2.T_DOUT_13): TriggerChannel.T_DOUT_13,
    int(allegoserver_pb2.T_DOUT_14): TriggerChannel.T_DOUT_14,
    int(allegoserver_pb2.T_DOUT_15): TriggerChannel.T_DOUT_15,
    int(allegoserver_pb2.T_NONE): TriggerChannel.T_NONE,
}


def trigger_channel_to_proto(tc: TriggerChannel) -> allegoserver_pb2.TriggerChannel.ValueType:
    """Convert domain TriggerChannel to proto TriggerChannel."""
    return _TRIGGER_CHANNEL_TO_PROTO[tc]


def trigger_channel_from_proto(pb: int) -> TriggerChannel:
    """Convert proto TriggerChannel int to domain TriggerChannel."""
    return _TRIGGER_CHANNEL_FROM_PROTO[pb]


# --- StimStep ---

_STIM_STEP_TO_PROTO: dict[StimStep, allegoserver_pb2.StimStep.ValueType] = {
    StimStep.STIM_STEP_SIZE_MIN: allegoserver_pb2.StimStepSizeMin,
    StimStep.STIM_STEP_SIZE_10_NA: allegoserver_pb2.StimStepSize10nA,
    StimStep.STIM_STEP_SIZE_20_NA: allegoserver_pb2.StimStepSize20nA,
    StimStep.STIM_STEP_SIZE_50_NA: allegoserver_pb2.StimStepSize50nA,
    StimStep.STIM_STEP_SIZE_100_NA: allegoserver_pb2.StimStepSize100nA,
    StimStep.STIM_STEP_SIZE_200_NA: allegoserver_pb2.StimStepSize200nA,
    StimStep.STIM_STEP_SIZE_500_NA: allegoserver_pb2.StimStepSize500nA,
    StimStep.STIM_STEP_SIZE_1_UA: allegoserver_pb2.StimStepSize1uA,
    StimStep.STIM_STEP_SIZE_2_UA: allegoserver_pb2.StimStepSize2uA,
    StimStep.STIM_STEP_SIZE_5_UA: allegoserver_pb2.StimStepSize5uA,
    StimStep.STIM_STEP_SIZE_10_UA: allegoserver_pb2.StimStepSize10uA,
    StimStep.STIM_STEP_SIZE_MAX: allegoserver_pb2.StimStepSizeMax,
}

_STIM_STEP_FROM_PROTO: dict[int, StimStep] = {
    int(allegoserver_pb2.StimStepSizeMin): StimStep.STIM_STEP_SIZE_MIN,
    int(allegoserver_pb2.StimStepSize10nA): StimStep.STIM_STEP_SIZE_10_NA,
    int(allegoserver_pb2.StimStepSize20nA): StimStep.STIM_STEP_SIZE_20_NA,
    int(allegoserver_pb2.StimStepSize50nA): StimStep.STIM_STEP_SIZE_50_NA,
    int(allegoserver_pb2.StimStepSize100nA): StimStep.STIM_STEP_SIZE_100_NA,
    int(allegoserver_pb2.StimStepSize200nA): StimStep.STIM_STEP_SIZE_200_NA,
    int(allegoserver_pb2.StimStepSize500nA): StimStep.STIM_STEP_SIZE_500_NA,
    int(allegoserver_pb2.StimStepSize1uA): StimStep.STIM_STEP_SIZE_1_UA,
    int(allegoserver_pb2.StimStepSize2uA): StimStep.STIM_STEP_SIZE_2_UA,
    int(allegoserver_pb2.StimStepSize5uA): StimStep.STIM_STEP_SIZE_5_UA,
    int(allegoserver_pb2.StimStepSize10uA): StimStep.STIM_STEP_SIZE_10_UA,
    int(allegoserver_pb2.StimStepSizeMax): StimStep.STIM_STEP_SIZE_MAX,
}


def stim_step_to_proto(ss: StimStep) -> allegoserver_pb2.StimStep.ValueType:
    """Convert domain StimStep to proto StimStep."""
    return _STIM_STEP_TO_PROTO[ss]


def stim_step_from_proto(pb: int) -> StimStep:
    """Convert proto StimStep int to domain StimStep."""
    return _STIM_STEP_FROM_PROTO[pb]


# --- DataRecordingType ---

_DATA_RECORDING_TYPE_TO_PROTO: dict[DataRecordingType, allegoserver_pb2.DataRecordingType.ValueType] = {
    DataRecordingType.CONTINUOUS: allegoserver_pb2.continuous,
    DataRecordingType.SPIKES: allegoserver_pb2.spikes,
    DataRecordingType.CONTINUOUS_AND_SPIKES: allegoserver_pb2.continuousAndSpikes,
}

_DATA_RECORDING_TYPE_FROM_PROTO: dict[int, DataRecordingType] = {
    int(allegoserver_pb2.continuous): DataRecordingType.CONTINUOUS,
    int(allegoserver_pb2.spikes): DataRecordingType.SPIKES,
    int(allegoserver_pb2.continuousAndSpikes): DataRecordingType.CONTINUOUS_AND_SPIKES,
}


def data_recording_type_to_proto(drt: DataRecordingType) -> allegoserver_pb2.DataRecordingType.ValueType:
    """Convert domain DataRecordingType to proto DataRecordingType."""
    return _DATA_RECORDING_TYPE_TO_PROTO[drt]


def data_recording_type_from_proto(pb: int) -> DataRecordingType:
    """Convert proto DataRecordingType int to domain DataRecordingType."""
    return _DATA_RECORDING_TYPE_FROM_PROTO[pb]



"""Stimulation step control converters — domain ↔ proto."""

from __future__ import annotations

from typing import TYPE_CHECKING

from radiens_core._converters._enums import (
    stim_step_from_proto,
    stim_step_to_proto,
)
from radiens_core.models.status import (
    StimStepRequest,
    StimStepState,
)

if TYPE_CHECKING:
    from radiens_core._generated import allegoserver_pb2


def stim_step_request_to_proto(
    request: StimStepRequest,
) -> allegoserver_pb2.StimStepMessage:
    """Convert ``StimStepRequest`` domain model → proto message."""
    from radiens_core._generated import allegoserver_pb2

    return allegoserver_pb2.StimStepMessage(
        stimStep=stim_step_to_proto(request.stim_step),
    )


def stim_step_state_from_proto(
    response: allegoserver_pb2.StimStepMessage,
) -> StimStepState:
    """Convert ``StimStepMessage`` proto response → domain model."""
    return StimStepState(
        stim_step=stim_step_from_proto(response.stimStep),
    )



"""DatasetMetadata — immutable metadata for a linked dataset."""

from pydantic import BaseModel, ConfigDict

from radiens_core.models.channels import ChannelMetadata
from radiens_core.models.enums import RadiensFileType
from radiens_core.models.time_range import TimeRange


class DatasetMetadata(BaseModel):
    """Metadata for a linked dataset.

    Returned by ``link_data_file()`` and similar operations.
    Contains everything needed to query signals from this dataset.
    """

    model_config = ConfigDict(frozen=True)

    id: str
    base_name: str
    path: str
    file_type: RadiensFileType
    time_range: TimeRange
    channel_metadata: ChannelMetadata
    parent_dsource_id: str = ""
    probe_uid: str = ""

    @property
    def duration(self) -> float:
        """Dataset duration in seconds."""
        return self.time_range.dur_sec

    @property
    def sampling_rate(self) -> float:
        """Sampling rate in Hz."""
        return self.time_range.fs



"""TimeRange — resolved, concrete time range with all fields populated."""

from pydantic import BaseModel, ConfigDict


class TimeRange(BaseModel):
    """A resolved, concrete time range.

    All fields are populated after resolution from a TimeRangeSpec
    against dataset metadata or cache state.

    Attributes
    ----------
    sec : tuple[float, float]
        Start and end times in seconds ``(start, end)``.
    timestamp : tuple[int, int]
        Start and end times in samples ``(start_sample, end_sample)``.
    fs : float
        Sampling rate in Hz.
    dur_sec : float
        Duration in seconds.
    walltime : str
        Wall-clock time string.
    n : int
        Number of samples.
    """

    model_config = ConfigDict(frozen=True)

    sec: tuple[float, float]
    timestamp: tuple[int, int]
    fs: float
    dur_sec: float
    walltime: str
    n: int



"""GPIO domain models."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class GPIOChannelCount(BaseModel):
    """GPIO channel count information."""

    model_config = ConfigDict(frozen=True)

    n_aux: int
    n_din: int
    n_dout: int



"""User-facing specification types — TimeRangeSpec, SignalSpec, Flex aliases.

These provide a clean, semantic interface for specifying time ranges
and signal selections across all client types.

Flex type aliases (``FlexTimeRange``, ``FlexSignalSpec``) use Pydantic
``BeforeValidator`` to coerce raw Python types (list, tuple, dict) into
canonical specs at the Pydantic model boundary.
"""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Annotated

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, BeforeValidator, ConfigDict

from radiens_core.models.enums import KeyIndex, SignalType, TrsMode
from radiens_core.models.signals import SigSelect

if TYPE_CHECKING:
    from radiens_core.models.channels import ChannelMetadata
    from radiens_core.models.time_range import TimeRange


class TimeRangeSpec(BaseModel):
    """Semantic time range specification with embedded mode.

    Use factory methods to create specs with clear semantics:

    >>> spec = TimeRangeSpec.subset(0, 10)     # [0, 10) seconds
    >>> spec = TimeRangeSpec.lookback(5.0)      # Last 5 seconds
    >>> spec = TimeRangeSpec.full()              # Entire recording
    """

    model_config = ConfigDict(frozen=True)

    mode: TrsMode
    start: float
    end: float | None = None

    # --- Factory methods ---

    @classmethod
    def subset(cls, start: float, end: float) -> TimeRangeSpec:
        """Absolute time range ``[start, end)`` seconds."""
        if end <= start:
            raise ValueError(f"end ({end}) must be greater than start ({start})")
        return cls(mode=TrsMode.SUBSET, start=start, end=end)

    @classmethod
    def lookback(cls, duration: float) -> TimeRangeSpec:
        """Last ``duration`` seconds from head of cache."""
        if duration <= 0:
            raise ValueError(f"duration must be positive, got {duration}")
        return cls(mode=TrsMode.FROM_HEAD, start=duration, end=None)

    @classmethod
    def to_head(cls, start: float) -> TimeRangeSpec:
        """From ``start`` seconds to current head."""
        return cls(mode=TrsMode.TO_HEAD, start=start, end=None)

    @classmethod
    def full(cls) -> TimeRangeSpec:
        """Full dataset/cache range."""
        return cls(mode=TrsMode.SUBSET, start=0, end=float("inf"))

    @classmethod
    def from_list(cls, time_range: list[float], mode: TrsMode | None = None) -> TimeRangeSpec:
        """Create from ``[start, end]`` or ``[duration, nan]`` list."""
        arr = np.asarray(time_range, dtype=np.float64)
        if len(arr) != 2:
            raise ValueError("time_range must have exactly 2 elements")
        if mode is None:
            mode = TrsMode.FROM_HEAD if np.isnan(arr[1]) else TrsMode.SUBSET
        if mode == TrsMode.FROM_HEAD:
            return cls(mode=mode, start=float(arr[0]), end=None)
        return cls(mode=mode, start=float(arr[0]), end=float(arr[1]))

    # --- Resolution methods ---

    def is_full(self) -> bool:
        """Check if this is a full-range specification."""
        return self.mode == TrsMode.SUBSET and self.end == float("inf")

    def to_array(self) -> npt.NDArray[np.float64]:
        """Convert to ``[start, end]`` or ``[duration, nan]`` array."""
        if self.is_full():
            raise ValueError("Cannot convert full() spec to array without dataset.")
        if self.mode in (TrsMode.FROM_HEAD, TrsMode.TO_HEAD):
            return np.array([self.start, np.nan], dtype=np.float64)
        return np.array([self.start, self.end], dtype=np.float64)

    def resolve(self, dataset_metadata: object | None = None, fs: float | None = None) -> TimeRange | None:
        """Resolve to a concrete TimeRange.

        Parameters
        ----------
        dataset_metadata
            Dataset metadata (must have ``.time_range`` and ``.make_time_range()``).
        fs
            Sampling rate override.

        Returns
        -------
        TimeRange or None
            None for FROM_HEAD mode (caller should use ``to_array()``).
        """
        from radiens_core.models.dataset import DatasetMetadata

        if self.is_full():
            if not isinstance(dataset_metadata, DatasetMetadata):
                raise ValueError("dataset_metadata required to resolve full() specification")
            return dataset_metadata.time_range

        if self.mode == TrsMode.FROM_HEAD:
            return None

        if isinstance(dataset_metadata, DatasetMetadata):
            return dataset_metadata.time_range

        raise ValueError("dataset_metadata required to fully resolve TimeRange")

    @property
    def duration(self) -> float | None:
        """Duration in seconds, if known."""
        if self.mode == TrsMode.FROM_HEAD:
            return self.start
        if self.mode == TrsMode.SUBSET and self.end is not None and self.end != float("inf"):
            return self.end - self.start
        return None

    def __repr__(self) -> str:
        if self.is_full():
            return "TimeRangeSpec.full()"
        if self.mode == TrsMode.FROM_HEAD:
            return f"TimeRangeSpec.lookback({self.start})"
        if self.mode == TrsMode.TO_HEAD:
            return f"TimeRangeSpec.to_head({self.start})"
        return f"TimeRangeSpec.subset({self.start}, {self.end})"


class SignalSpec(BaseModel):
    """Semantic signal selection specification.

    Use factory methods for clean channel selection:

    >>> spec = SignalSpec.amp([0, 1, 2, 3])
    >>> spec = SignalSpec.ain("all")
    >>> spec = SignalSpec.multi(amp=[0, 1], din="all")
    """

    model_config = ConfigDict(frozen=True)

    sig_type: SignalType | None
    channels: tuple[int, ...] | str
    multi_select: dict[SignalType, tuple[int, ...] | str] | None = None

    # --- Single-type factory methods ---

    @classmethod
    def amp(cls, channels: list[int] | str = "all") -> SignalSpec:
        """Select amplifier channels."""
        return cls._single(SignalType.AMP, channels)

    @classmethod
    def ain(cls, channels: list[int] | str = "all") -> SignalSpec:
        """Select analog input channels."""
        return cls._single(SignalType.AIN, channels)

    @classmethod
    def din(cls, channels: list[int] | str = "all") -> SignalSpec:
        """Select digital input channels."""
        return cls._single(SignalType.DIN, channels)

    @classmethod
    def dout(cls, channels: list[int] | str = "all") -> SignalSpec:
        """Select digital output channels."""
        return cls._single(SignalType.DOUT, channels)

    @classmethod
    def all_amp(cls) -> SignalSpec:
        """Convenience: all amplifier channels."""
        return cls.amp("all")

    @classmethod
    def _single(cls, sig_type: SignalType, channels: list[int] | str) -> SignalSpec:
        ch: tuple[int, ...] | str = tuple(channels) if isinstance(channels, list) else channels
        return cls(sig_type=sig_type, channels=ch, multi_select=None)

    # --- Multi-type factory method ---

    @classmethod
    def multi(
        cls,
        amp: list[int] | str | None = None,
        ain: list[int] | str | None = None,
        din: list[int] | str | None = None,
        dout: list[int] | str | None = None,
    ) -> SignalSpec:
        """Select channels from multiple signal types."""
        multi_dict: dict[SignalType, tuple[int, ...] | str] = {}
        for sig_type, ch in [
            (SignalType.AMP, amp),
            (SignalType.AIN, ain),
            (SignalType.DIN, din),
            (SignalType.DOUT, dout),
        ]:
            if ch is not None:
                multi_dict[sig_type] = tuple(ch) if isinstance(ch, list) else ch
        if not multi_dict:
            raise ValueError("At least one signal type must be specified")
        return cls(sig_type=None, channels="multi", multi_select=multi_dict)

    # --- Resolution ---

    def is_multi(self) -> bool:
        """Check if this is a multi-type specification."""
        return self.multi_select is not None

    def to_sig_select(self, channel_metadata: ChannelMetadata) -> SigSelect:
        """Convert to SigSelect for API calls."""
        if self.is_multi():
            return self._resolve_multi(channel_metadata)
        return self._resolve_single(channel_metadata)

    def _resolve_single(self, channel_metadata: ChannelMetadata) -> SigSelect:
        arrays: dict[SignalType, npt.NDArray[np.int64]] = {
            st: np.array([], dtype=np.int64) for st in SignalType
        }
        if self.sig_type is not None:
            arrays[self.sig_type] = self._resolve_channels(
                self.sig_type, self.channels, channel_metadata
            )
        return SigSelect(
            key_idx=KeyIndex.NTV,
            amp=arrays[SignalType.AMP],
            gpio_ain=arrays[SignalType.AIN],
            gpio_din=arrays[SignalType.DIN],
            gpio_dout=arrays[SignalType.DOUT],
        )

    def _resolve_multi(self, channel_metadata: ChannelMetadata) -> SigSelect:
        arrays: dict[SignalType, npt.NDArray[np.int64]] = {
            st: np.array([], dtype=np.int64) for st in SignalType
        }
        if self.multi_select is not None:
            for sig_type, channels in self.multi_select.items():
                arrays[sig_type] = self._resolve_channels(sig_type, channels, channel_metadata)
        return SigSelect(
            key_idx=KeyIndex.NTV,
            amp=arrays[SignalType.AMP],
            gpio_ain=arrays[SignalType.AIN],
            gpio_din=arrays[SignalType.DIN],
            gpio_dout=arrays[SignalType.DOUT],
        )

    @staticmethod
    def _resolve_channels(
        sig_type: SignalType,
        channels: tuple[int, ...] | str,
        channel_metadata: ChannelMetadata,
    ) -> npt.NDArray[np.int64]:
        available = channel_metadata.index(sig_type).ntv
        if channels == "all":
            return np.array(available, dtype=np.int64)
        if channels == "selected":
            selected = channel_metadata.selected_index(sig_type).ntv
            return np.array(selected, dtype=np.int64)
        if isinstance(channels, tuple):
            invalid = set(channels) - set(available)
            if invalid:
                raise ValueError(
                    f"Invalid {sig_type.name.lower()} channel indices: {invalid}. "
                    f"Available: {available}"
                )
            return np.array(channels, dtype=np.int64)
        raise ValueError(f"Invalid channel specification: {channels}")

    def __repr__(self) -> str:
        if self.is_multi() and self.multi_select is not None:
            parts: list[str] = []
            for sig_type, channels in self.multi_select.items():
                ch_display = list(channels) if isinstance(channels, tuple) else repr(channels)
                parts.append(f"{sig_type.name.lower()}={ch_display}")
            return f"SignalSpec.multi({', '.join(parts)})"
        ch_display_single = list(self.channels) if isinstance(self.channels, tuple) else repr(self.channels)
        name = self.sig_type.name.lower() if self.sig_type is not None else "unknown"
        return f"SignalSpec.{name}({ch_display_single})"


# ===================================================================
# Flex type aliases — Pydantic BeforeValidator coercions
# ===================================================================


def _coerce_time_range(v: object) -> TimeRangeSpec:
    """Coerce raw Python types into :class:`TimeRangeSpec`.

    Accepted inputs:
    - ``TimeRangeSpec`` — pass-through
    - ``list[float]`` / ``tuple[float, ...]`` — delegates to ``TimeRangeSpec.from_list``
    """
    if isinstance(v, TimeRangeSpec):
        return v
    if isinstance(v, (list, tuple)):
        return TimeRangeSpec.from_list(list(v))
    raise ValueError(f"Expected TimeRangeSpec, list, or tuple — got {type(v).__name__}")


def _coerce_signal_spec(v: object) -> SignalSpec:
    """Coerce raw Python types into :class:`SignalSpec`.

    Accepted inputs:
    - ``SignalSpec`` — pass-through
    - ``list[int]`` — interpreted as amplifier channel indices
    - ``dict`` — passed as kwargs to ``SignalSpec.multi()``
    """
    if isinstance(v, SignalSpec):
        return v
    if isinstance(v, list):
        return SignalSpec.amp(v)
    if isinstance(v, dict):
        return SignalSpec.multi(**v)
    raise ValueError(f"Expected SignalSpec, list[int], or dict — got {type(v).__name__}")


FlexTimeRange = Annotated[
    TimeRangeSpec | Sequence[float],
    BeforeValidator(_coerce_time_range),
]
"""``TimeRangeSpec`` that also accepts ``list``/``tuple`` shorthand.

Use in Pydantic models::

    class MyRequest(BaseModel):
        time_range: FlexTimeRange
"""

FlexSignalSpec = Annotated[
    SignalSpec | Sequence[int] | Mapping[str, Sequence[int] | str],
    BeforeValidator(_coerce_signal_spec),
]
"""``SignalSpec`` that also accepts ``list[int]`` or ``dict`` shorthand.

Use in Pydantic models::

    class MyRequest(BaseModel):
        signals: FlexSignalSpec
"""



"""Stimulation parameter converter — proto ↔ domain."""

from __future__ import annotations

from typing import TYPE_CHECKING

from radiens_core._converters._enums import (
    stim_polarity_from_proto,
    stim_polarity_to_proto,
    stim_pulse_mode_from_proto,
    stim_pulse_mode_to_proto,
    stim_shape_from_proto,
    stim_shape_to_proto,
    stim_trigger_edge_or_level_from_proto,
    stim_trigger_edge_or_level_to_proto,
    stim_trigger_level_from_proto,
    stim_trigger_level_to_proto,
)
from radiens_core.models.stim import StimParams

if TYPE_CHECKING:
    from radiens_core._generated import allegoserver_pb2


def stim_params_from_proto(
    reply: allegoserver_pb2.StimParamsReply,
) -> dict[int, StimParams]:
    """Convert ``StimParamsReply`` proto → dict keyed by system channel index."""
    result: dict[int, StimParams] = {}
    for sys_chan_idx, pb in reply.params.items():
        result[sys_chan_idx] = _single_from_proto(pb)
    return result


def _single_from_proto(pb: allegoserver_pb2.StimParams) -> StimParams:
    """Convert a single ``StimParams`` proto message to a domain model."""
    return StimParams(
        stim_shape=stim_shape_from_proto(pb.stimShape),
        stim_polarity=stim_polarity_from_proto(pb.stimPolarity),
        first_phase_duration=pb.firstPhaseDuration,
        second_phase_duration=pb.secondPhaseDuration,
        interphase_delay=pb.interphaseDelay,
        first_phase_amplitude=pb.firstPhaseAmplitude,
        second_phase_amplitude=pb.secondPhaseAmplitude,
        baseline_voltage=pb.baselineVoltage,
        trigger_edge_or_level=stim_trigger_edge_or_level_from_proto(pb.triggerEdgeOrLevel),
        trigger_high_or_low=stim_trigger_level_from_proto(pb.triggerHighOrLow),
        enabled=pb.enabled,
        post_trigger_delay=pb.postTriggerDelay,
        pulse_or_train=stim_pulse_mode_from_proto(pb.pulseOrTrain),
        number_of_stim_pulses=pb.numberOfStimPulses,
        pulse_train_period=pb.pulseTrainPeriod,
        refractory_period=pb.refractoryPeriod,
        pre_stim_amp_settle=pb.preStimAmpSettle,
        post_stim_amp_settle=pb.postStimAmpSettle,
        maintain_amp_settle=pb.maintainAmpSettle,
        enable_amp_settle=pb.enableAmpSettle,
        post_stim_charge_recov_on=pb.postStimChargeRecovOn,
        post_stim_charge_recov_off=pb.postStimChargeRecovOff,
        enable_charge_recovery=pb.enableChargeRecovery,
        trigger_source_is_keypress=pb.triggerSourceIsKeypress,
        trigger_source_idx=pb.triggerSourceIdx,
        stim_sys_chan_idx=pb.stimSysChanIdx,
    )


def stim_params_to_proto(
    stim_params: StimParams,
) -> allegoserver_pb2.StimParams:
    """Convert ``StimParams`` domain model → proto message."""
    from radiens_core._generated import allegoserver_pb2

    return allegoserver_pb2.StimParams(
        stimShape=stim_shape_to_proto(stim_params.stim_shape),
        stimPolarity=stim_polarity_to_proto(stim_params.stim_polarity),
        firstPhaseDuration=stim_params.first_phase_duration,
        secondPhaseDuration=stim_params.second_phase_duration,
        interphaseDelay=stim_params.interphase_delay,
        firstPhaseAmplitude=stim_params.first_phase_amplitude,
        secondPhaseAmplitude=stim_params.second_phase_amplitude,
        baselineVoltage=stim_params.baseline_voltage,
        triggerEdgeOrLevel=stim_trigger_edge_or_level_to_proto(
            stim_params.trigger_edge_or_level
        ),
        triggerHighOrLow=stim_trigger_level_to_proto(
            stim_params.trigger_high_or_low
        ),
        enabled=stim_params.enabled,
        postTriggerDelay=stim_params.post_trigger_delay,
        pulseOrTrain=stim_pulse_mode_to_proto(stim_params.pulse_or_train),
        numberOfStimPulses=stim_params.number_of_stim_pulses,
        pulseTrainPeriod=stim_params.pulse_train_period,
        refractoryPeriod=stim_params.refractory_period,
        preStimAmpSettle=stim_params.pre_stim_amp_settle,
        postStimAmpSettle=stim_params.post_stim_amp_settle,
        maintainAmpSettle=stim_params.maintain_amp_settle,
        enableAmpSettle=stim_params.enable_amp_settle,
        postStimChargeRecovOn=stim_params.post_stim_charge_recov_on,
        postStimChargeRecovOff=stim_params.post_stim_charge_recov_off,
        enableChargeRecovery=stim_params.enable_charge_recovery,
        triggerSourceIsKeypress=stim_params.trigger_source_is_keypress,
        triggerSourceIdx=stim_params.trigger_source_idx,
        stimSysChanIdx=stim_params.stim_sys_chan_idx,
    )



"""System status domain models."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict

from radiens_core.models.enums import (
    BackboneMode,
    DataRecordingType,
    Port,
    RecordMode,
    StimKeypressIndex,
    StimStep,
    StreamMode,
    TriggerChannel,
)
from radiens_core.models.gpio import GPIOChannelCount
from radiens_core.models.ports import PortChannelCount


class RestartRequest(BaseModel):
    """Request to restart the system with a specific backbone mode."""

    model_config = ConfigDict(frozen=True)

    mode: BackboneMode


class RecordStateRequest(BaseModel):
    """Request to change recording state."""

    model_config = ConfigDict(frozen=True)

    mode: RecordMode
    port: Port | None = None  # If None, system-level record control


class StreamStateRequest(BaseModel):
    """Request to change streaming state."""

    model_config = ConfigDict(frozen=True)

    mode: StreamMode


class StreamingStatus(BaseModel):
    """Current streaming status."""

    model_config = ConfigDict(frozen=True)

    stream_mode: StreamMode
    primary_cache_t_range: list[float]
    hardware_memory_level: float


class RecordingStatus(BaseModel):
    """Current recording status."""

    model_config = ConfigDict(frozen=True)

    record_mode: RecordMode
    active_filename: str
    duration: float
    error: str
    record_mode_port_a: RecordMode | None = None
    record_mode_port_b: RecordMode | None = None
    record_mode_port_c: RecordMode | None = None
    record_mode_port_d: RecordMode | None = None
    record_mode_port_e: RecordMode | None = None
    record_mode_port_f: RecordMode | None = None
    record_mode_port_g: RecordMode | None = None
    record_mode_port_h: RecordMode | None = None


class RecordingConfig(BaseModel):
    """Recording configuration settings.

    Configures output file paths, naming, and recording behavior.
    Used as both request (for set_recording_config) and response (for get_recording_config).
    """

    model_config = ConfigDict(frozen=True)

    base_file_name: str
    """Base filename for recording output (without extension or timestamp)."""

    base_file_path: str
    """Base directory path for recording files."""

    datasource_idx: int = 0
    """Index number for the composed file name."""

    time_stamp: bool = True
    """True to include timestamp in the composed file name."""

    split_files_by_port: bool = False
    """True to create separate files for each port."""

    force_start_at_tstamp_0: bool = False
    """True to force recording to start at timestamp 0."""

    recording_type: DataRecordingType = DataRecordingType.CONTINUOUS
    """Type of data to record (continuous, spikes, or both)."""


class BackboneStatus(BaseModel):
    """Complete system backbone status."""

    model_config = ConfigDict(frozen=True)

    streaming: StreamingStatus
    recording: RecordingStatus
    ports: list[PortChannelCount]
    is_connected: bool
    gpio_channel_count: GPIOChannelCount


class SetTriggerStateRequest(BaseModel):
    """Request to set trigger channel state.

    Configures which trigger channels are enabled for stimulation.
    """

    model_config = ConfigDict(frozen=True)

    channel: TriggerChannel
    enabled: bool
    port: Port


class TriggerState(BaseModel):
    """Current trigger channel configuration.

    Shows which channels and ports are currently enabled for triggering.
    """

    model_config = ConfigDict(frozen=True)

    enabled_channels: list[TriggerChannel]
    ports: list[Port]


class ManualStimTriggerRequest(BaseModel):
    """Request for manual stimulation trigger.

    Allows direct control of stimulation triggering.
    """

    model_config = ConfigDict(frozen=True)

    trigger: StimKeypressIndex
    """Keypress trigger (1-8). Matches ``StimParams.trigger_source_idx``."""
    trigger_on: bool


class ManualStimTriggerToggleRequest(BaseModel):
    """Request to toggle manual stimulation trigger.

    Convenient toggle interface for stimulation control.
    """

    model_config = ConfigDict(frozen=True)

    trigger: StimKeypressIndex
    """Keypress trigger (1-8). Matches ``StimParams.trigger_source_idx``."""


class StimStepRequest(BaseModel):
    """Request to set stimulation step size.

    Configures the current increment/decrement granularity.
    """

    model_config = ConfigDict(frozen=True)

    stim_step: StimStep


class StimStepState(BaseModel):
    """Current stimulation step size setting.

    Shows the current stimulation step configuration.
    """

    model_config = ConfigDict(frozen=True)

    stim_step: StimStep



"""Domain enums — proto-free, plain Python IntEnum/StrEnum values.

These enums mirror the proto enum values but have ZERO dependency on generated code.
The mapping between domain enums and proto enums lives in ``_converters/_enums.py``.
"""

from enum import IntEnum, StrEnum

__all__ = [
    "AreaUnits",
    "BackboneMode",
    "ClientType",
    "DataRecordingType",
    "KeyIndex",
    "Port",
    "RadiensFileType",
    "RadiensService",
    "RecordMode",
    "SignalType",
    "SignalUnits",
    "StimPolarity",
    "StimPulseMode",
    "StimShape",
    "StimTriggerEdgeOrLevel",
    "StimTriggerHighOrLow",
    "StreamMode",
    "TrsMode",
]


class ClientType(StrEnum):
    """Client type identifier."""

    ALLEGO = "AllegoClient"
    VIDERE = "VidereClient"
    CURATE = "CurateClient"


class RadiensService(IntEnum):
    """Radiens backend service type."""

    CORE = 0
    DASH = 1
    SORTER = 2
    DEV = 3
    RADIENSPY = 4
    PCACHE = 5


class KeyIndex(IntEnum):
    """Channel key index type.

    NTV: native position in DAQ stream per signal type (starts at 0).
    DATA: position in backing dataset per signal type.
    SYS: global position across all signal types.
    """

    NTV = 0
    DATA = 1
    SYS = 2


class SignalType(IntEnum):
    """Signal type. Values match proto ``common.SignalType``."""

    AMP = 0   # PRI in proto
    AIN = 1   # AUX in proto
    DIN = 2
    DOUT = 3


class TrsMode(IntEnum):
    """Time range selection mode. Values match proto ``common.TimeRangeSelMode``."""

    SUBSET = 0     # TRS_SUBSET
    TO_HEAD = 1    # TRS_TO_HEAD
    FROM_HEAD = 2  # TRS_FROM_HEAD


class RadiensFileType(IntEnum):
    """Recording file type. Values match proto ``common.RadixFileTypes``."""

    RHD = 0
    XDAT = 1
    CSV = 2
    HDF5 = 3
    NEX5 = 4
    NWB = 5
    KILOSORT2 = 8
    NSX = 13
    TDT = 15
    SPIKES = 16
    OE_DAT = 22


class SignalUnits(IntEnum):
    """Signal measurement units."""

    UNKNOWN = 0
    MICROVOLTS = 1
    VOLTS = 2
    BINARY = 3


class AreaUnits(IntEnum):
    """Area measurement units for electrode sites."""

    SQUARE_MICROMETERS = 0  # µm² (base unit)
    SQUARE_MILLIMETERS = 1  # mm²
    SQUARE_CENTIMETERS = 2  # cm²
    SQUARE_NANOMETERS = 3  # nm²
    SQUARE_METERS = 4  # m²


# ===================================================================
# Stimulation enums
# ===================================================================


class StimShape(IntEnum):
    """Stimulation waveform shape. Values match proto ``allego.StimShape``."""

    BIPHASIC = 0
    BIPHASIC_INTERPHASE_DELAY = 1
    TRIPHASIC = 2
    MONOPHASIC = 3


class StimPolarity(IntEnum):
    """Stimulation polarity. Values match proto ``allego.StimPolarity``."""

    CATHODIC_FIRST = 0
    ANODIC_FIRST = 1


class StimTriggerEdgeOrLevel(IntEnum):
    """Stimulation trigger edge/level mode. Values match proto ``allego.StimTriggerEdgeOrLevel``."""

    EDGE = 0
    LEVEL = 1


class StimTriggerHighOrLow(IntEnum):
    """Stimulation trigger high/low. Values match proto ``allego.StimTriggerHighOrLow``."""

    HIGH = 0
    LOW = 1


class StimPulseMode(IntEnum):
    """Stimulation pulse mode. Values match proto ``allego.StimPulseOrTrain``."""

    SINGLE_PULSE = 0
    PULSE_TRAIN = 1


# ===================================================================
# System status enums
# ===================================================================


class BackboneMode(IntEnum):
    """Hardware backbone mode. Values match proto ``allego.BackboneMode``."""

    SMARTBOX_PRO = 0
    SMARTBOX_SIM_GEN_SINE = 1  # low freq
    SMARTBOX_SIM_GEN_SPIKES = 2
    SMARTBOX_SIM_DATASOURCE = 3
    OPEN_EPHYS_USB3 = 4
    INTAN_RECORDING_CONTROLLER_1024 = 5
    SMARTBOX_CLASSIC = 6
    INTAN_RECORDING_CONTROLLER_512 = 7
    OPEN_EPHYS_USB2 = 8
    INTAN_USB2 = 9
    SMARTBOX_SIM_GEN_SINE_MAPPED = 10
    SMARTBOX_SIM_GEN_SINE_HIGH_FREQ = 11
    SMARTBOX_SIM_GEN_SINE_MULTI_BAND = 12
    SMARTBOX_PRO_SINAPS_256 = 13
    SMARTBOX_PRO_SINAPS_1024 = 14
    XDAQ_ONE_REC = 15
    XDAQ_ONE_STIM = 16
    XDAQ_CORE_REC = 17
    XDAQ_CORE_STIM = 18
    SMARTBOX_PRO_SINAPS_512 = 19
    XDAQ_GEN2_ONE_STIM = 20
    XDAQ_GEN2_CORE_STIM = 21
    XDAQ_GEN2_ONE_REC = 22
    XDAQ_GEN2_CORE_REC = 23
    SMARTBOX_SIM_GEN_LFP = 24


class StreamMode(IntEnum):
    """Data stream mode. Values match proto ``allego.StreamMode``."""

    OFF = 0  # S_OFF
    ON = 1   # S_ON


class RecordMode(IntEnum):
    """Data recording mode. Values match proto ``allego.RecordMode``."""

    OFF = 0  # R_OFF
    ON = 1   # R_ON


class DataRecordingType(IntEnum):
    """Data recording type. Values match proto ``allego.DataRecordingType``."""

    CONTINUOUS = 0           # continuous
    SPIKES = 1              # spikes
    CONTINUOUS_AND_SPIKES = 2    # continuousAndSpikes


class Port(IntEnum):
    """Hardware port identifier. Values match proto ``common.Port``."""

    A = 0
    B = 1
    C = 2
    D = 3
    E = 4
    F = 5
    G = 6
    H = 7


class TriggerChannel(IntEnum):
    """Trigger channel identifier. Values match proto ``allego.TriggerChannel``."""

    T_DIN_0 = 0
    T_DIN_1 = 1
    T_DIN_2 = 2
    T_DIN_3 = 3
    T_DIN_4 = 4
    T_DIN_5 = 5
    T_DIN_6 = 6
    T_DIN_7 = 7
    T_DIN_8 = 8
    T_DIN_9 = 9
    T_DIN_10 = 10
    T_DIN_11 = 11
    T_DIN_12 = 12
    T_DIN_13 = 13
    T_DIN_14 = 14
    T_DIN_15 = 15
    T_DOUT_0 = 16
    T_DOUT_1 = 17
    T_DOUT_2 = 18
    T_DOUT_3 = 19
    T_DOUT_4 = 20
    T_DOUT_5 = 21
    T_DOUT_6 = 22
    T_DOUT_7 = 23
    T_DOUT_8 = 24
    T_DOUT_9 = 25
    T_DOUT_10 = 26
    T_DOUT_11 = 27
    T_DOUT_12 = 28
    T_DOUT_13 = 29
    T_DOUT_14 = 30
    T_DOUT_15 = 31
    T_NONE = 32


class StimStep(IntEnum):
    """Stimulation step size. Values match proto ``allego.StimStep``."""

    STIM_STEP_SIZE_MIN = 0
    STIM_STEP_SIZE_10_NA = 1
    STIM_STEP_SIZE_20_NA = 2
    STIM_STEP_SIZE_50_NA = 3
    STIM_STEP_SIZE_100_NA = 4
    STIM_STEP_SIZE_200_NA = 5
    STIM_STEP_SIZE_500_NA = 6
    STIM_STEP_SIZE_1_UA = 7
    STIM_STEP_SIZE_2_UA = 8
    STIM_STEP_SIZE_5_UA = 9
    STIM_STEP_SIZE_10_UA = 10
    STIM_STEP_SIZE_MAX = 11


class StimKeypressIndex(IntEnum):
    KEYPRESS_1 = 1
    KEYPRESS_2 = 2
    KEYPRESS_3 = 3
    KEYPRESS_4 = 4
    KEYPRESS_5 = 5
    KEYPRESS_6 = 6
    KEYPRESS_7 = 7
    KEYPRESS_8 = 8



"""Radiens Python client for neuroscience data acquisition and analysis."""

from radiens_core._version import __version__
from radiens_core.allego_client import AllegoClient
from radiens_core.curate_client import CurateClient
from radiens_core.exceptions import (
    ConfigurationError,
    DataProcessingError,
    DatasetError,
    RadiensError,
    ServerCommunicationError,
    SignalSelectionError,
    TimeRangeError,
)
from radiens_core.models.channels import ChannelMetadata
from radiens_core.models.dataset import DatasetMetadata
from radiens_core.models.enums import (
    ClientType,
    DataRecordingType,
    KeyIndex,
    Port,
    RadiensFileType,
    RecordMode,
    SignalType,
    StimPolarity,
    StimPulseMode,
    StimShape,
    StimStep,
    StimTriggerEdgeOrLevel,
    StimTriggerHighOrLow,
    StreamMode,
    TrsMode,
)
from radiens_core.models.gpio import GPIOChannelCount
from radiens_core.models.ports import PortChannelCount
from radiens_core.models.signals import SignalArrays, SigSelect
from radiens_core.models.status import (
    BackboneStatus,
    RecordingConfig,
    RecordingStatus,
    StimStepRequest,
    StimStepState,
    StreamingStatus,
)
from radiens_core.models.stim import StimParams
from radiens_core.models.time_range import TimeRange
from radiens_core.specs import FlexSignalSpec, FlexTimeRange, SignalSpec, TimeRangeSpec
from radiens_core.videre_client import VidereClient

__all__ = [
    "AllegoClient",
    "BackboneStatus",
    "ChannelMetadata",
    "ClientType",
    "ConfigurationError",
    "CurateClient",
    "DataProcessingError",
    "DataRecordingType",
    "DatasetError",
    "DatasetMetadata",
    "FlexSignalSpec",
    "FlexTimeRange",
    "GPIOChannelCount",
    "KeyIndex",
    "Port",
    "PortChannelCount",
    "RadiensError",
    "RadiensFileType",
    "RecordMode",
    "RecordingConfig",
    "RecordingStatus",
    "ServerCommunicationError",
    "SigSelect",
    "SignalArrays",
    "SignalSelectionError",
    "SignalSpec",
    "SignalType",
    "StimParams",
    "StimPolarity",
    "StimPulseMode",
    "StimShape",
    "StimStep",
    "StimStepRequest",
    "StimStepState",
    "StimTriggerEdgeOrLevel",
    "StimTriggerHighOrLow",
    "StreamMode",
    "StreamingStatus",
    "TimeRange",
    "TimeRangeError",
    "TimeRangeSpec",
    "TrsMode",
    "VidereClient",
    "__version__",
]
